@@ -28,10 +28,7 @@ OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
-
-#ifdef HAVE_CONFIG_H
-#include "io_lib_config.h"
-#endif
+#include <config.h>
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -45,7 +42,10 @@ OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
 #include "cram/os.h"
 #include "cram/mFILE.h"
-#include "cram/vlen.h"
+
+#ifdef HAVE_MMAP
+#include <sys/mman.h>
+#endif
 
 /*
  * This file contains memory-based versions of the most commonly used
@@ -106,6 +106,33 @@ static char *mfload(FILE *fp, const char *fn, size_t *size, int binary) {
     return data;
 }
 
+
+#ifdef HAVE_MMAP
+/*
+ * mmaps in the file, but only for reading currently.
+ *
+ * Returns 0 on success
+ *        -1 on failure
+ */
+int mfmmap(mFILE *mf, FILE *fp, const char *fn) {
+    struct stat sb;
+
+    if (stat(fn, &sb) != 0)
+	return -1;
+
+    mf->size = sb.st_size;
+    mf->data = mmap(NULL, mf->size, PROT_READ, MAP_SHARED,
+		    fileno(fp), 0);
+
+    if (!mf->data)
+	return -1;
+
+    mf->alloced = 0;
+    return 0;
+}
+#endif
+
+
 /*
  * Creates and returns m_channel[0].
  * We initialise this on the first attempted read, which then slurps in
@@ -233,6 +260,8 @@ mFILE *mfreopen(const char *path, const char *mode_str, FILE *fp) {
      * w = write on close
      * a = position at end of buffer
      * x = position at same location as the original fp, don't seek on flush
+     * + = for update (read and write)
+     * m = mmap (read only)
      */
     if (strchr(mode_str, 'r'))
 	r = 1, mode |= MF_READ;
@@ -249,15 +278,29 @@ mFILE *mfreopen(const char *path, const char *mode_str, FILE *fp) {
 	if (a)
 	    r = 1;
     }
+#ifdef HAVE_MMAP
+    if (strchr(mode_str, 'm'))
+	if (!w) mode |= MF_MMAP;
+#endif
 
     if (r) {
 	mf = mfcreate(NULL, 0);
 	if (NULL == mf) return NULL;
 	if (!(mode & MF_TRUNC)) {
-	    mf->data = mfload(fp, path, &mf->size, b);
-	    mf->alloced = mf->size;
-	    if (!a)
-		fseek(fp, 0, SEEK_SET);
+#ifdef HAVE_MMAP
+	    if (mode & MF_MMAP) {
+		if (mfmmap(mf, fp, path) == -1) {
+		    mf->data = NULL;
+		    mode &= ~MF_MMAP;
+		}
+	    }
+#endif
+	    if (!mf->data) {
+		mf->data = mfload(fp, path, &mf->size, b);
+		mf->alloced = mf->size;
+		if (!a)
+		    fseek(fp, 0, SEEK_SET);
+	    }
 	}
     } else if (w) {
 	/* Write - initialise the data structures */
@@ -307,6 +350,14 @@ int mfclose(mFILE *mf) {
 
     mfflush(mf);
 
+#ifdef HAVE_MMAP
+    if ((mf->mode & MF_MMAP) && mf->data) {
+	/* Mmaped */
+	munmap(mf->data, mf->size);
+	mf->data = NULL;
+    }
+#endif
+
     if (mf->fp)
 	fclose(mf->fp);
 
@@ -318,12 +369,16 @@ int mfclose(mFILE *mf) {
 /*
  * Closes the file pointer contained within the mFILE without destroying
  * the in-memory data.
+ *
+ * Attempting to do this on an mmaped buffer is an error.
  */
 int mfdetach(mFILE *mf) {
     if (!mf)
 	return -1;
 
     mfflush(mf);
+    if (mf->mode & MF_MMAP)
+	return -1;
 
     if (mf->fp) {
 	fclose(mf->fp);
@@ -352,6 +407,8 @@ int mfdestroy(mFILE *mf) {
  * It is up to the caller to free the stolen buffer.  If size_out is
  * not NULL, mf->size will be stored in it.
  * This is more-or-less the opposite of mfcreate().
+ *
+ * Note, we cannot steal the allocated buffer from an mmaped mFILE.
  */
 
 void *mfsteal(mFILE *mf, size_t *size_out) {
@@ -363,7 +420,9 @@ void *mfsteal(mFILE *mf, size_t *size_out) {
     
     if (NULL != size_out) *size_out = mf->size;
 
-    mfdetach(mf);
+    if (mfdetach(mf) != 0)
+	return NULL;
+
     mf->data = NULL;
     mfdestroy(mf);
 
@@ -570,45 +629,6 @@ int mfflush(mFILE *mf) {
     return 0;
 }
 
-/*
- * A wrapper around vsprintf() to write to an mFILE. This also uses vflen() to
- * estimate how many additional bytes of storage will be required for the
- * vsprintf to work.
- */
-int mfprintf(mFILE *mf, char *fmt, ...) {
-    int ret;
-    size_t est_length;
-    va_list args;
-
-    va_start(args, fmt);
-    est_length = vflen(fmt, args);
-    va_end(args);
-    while (est_length + mf->offset > mf->alloced) {
-	size_t new_alloced = mf->alloced ? mf->alloced * 2 : 1024;
-	void * new_data    = realloc(mf->data, new_alloced);
-	if (NULL == new_data) return -1;
-	mf->alloced = new_alloced;
-	mf->data    = new_data;
-    }
-
-    va_start(args, fmt);
-    ret = vsprintf(&mf->data[mf->offset], fmt, args);
-    va_end(args);
-
-    if (ret > 0) {
-	mf->offset += ret;
-	if (mf->size < mf->offset)
-	    mf->size = mf->offset;
-    }
-
-    if (mf->fp == stderr) {
-	/* Auto-flush for stderr */
-	if (0 != mfflush(mf)) return -1;
-    }
-
-    return ret;
-}
-
 /*
  * Converts an mFILE from binary to ascii mode by replacing all
  * cr-nl with nl.
