@@ -1,6 +1,6 @@
 /*  faidx.c -- FASTA random access.
 
-    Copyright (C) 2008, 2009, 2013-2016 Genome Research Ltd.
+    Copyright (C) 2008, 2009, 2013-2017 Genome Research Ltd.
     Portions copyright (C) 2011 Broad Institute.
 
     Author: Heng Li <lh3@sanger.ac.uk>
@@ -31,12 +31,16 @@ DEALINGS IN THE SOFTWARE.  */
 #include <stdio.h>
 #include <inttypes.h>
 #include <errno.h>
+#include <limits.h>
+#include <unistd.h>
+#include <assert.h>
 
 #include "htslib/bgzf.h"
 #include "htslib/faidx.h"
 #include "htslib/hfile.h"
 #include "htslib/khash.h"
 #include "htslib/kstring.h"
+#include "hts_internal.h"
 
 typedef struct {
     int32_t line_len, line_blen;
@@ -56,10 +60,10 @@ struct __faidx_t {
 #define kroundup32(x) (--(x), (x)|=(x)>>1, (x)|=(x)>>2, (x)|=(x)>>4, (x)|=(x)>>8, (x)|=(x)>>16, ++(x))
 #endif
 
-static inline int fai_insert_index(faidx_t *idx, const char *name, int len, int line_len, int line_blen, uint64_t offset)
+static inline int fai_insert_index(faidx_t *idx, const char *name, int64_t len, int line_len, int line_blen, uint64_t offset)
 {
     if (!name) {
-        fprintf(stderr, "[fai_build_core] malformed line\n");
+        hts_log_error("Malformed line");
         return -1;
     }
 
@@ -69,7 +73,7 @@ static inline int fai_insert_index(faidx_t *idx, const char *name, int len, int
     faidx1_t *v = &kh_value(idx->hash, k);
 
     if (! absent) {
-        fprintf(stderr, "[fai_build_core] ignoring duplicate sequence \"%s\" at byte offset %"PRIu64"\n", name, offset);
+        hts_log_warning("Ignoring duplicate sequence \"%s\" at byte offset %"PRIu64"", name, offset);
         free(name_key);
         return 0;
     }
@@ -78,7 +82,7 @@ static inline int fai_insert_index(faidx_t *idx, const char *name, int len, int
         char **tmp;
         idx->m = idx->m? idx->m<<1 : 16;
         if (!(tmp = (char**)realloc(idx->name, sizeof(char*) * idx->m))) {
-            fprintf(stderr, "[fai_build_core] out of memory\n");
+            hts_log_error("Out of memory");
             return -1;
         }
         idx->name = tmp;
@@ -126,7 +130,7 @@ faidx_t *fai_build_core(BGZF *bgzf)
             kputsn("", 0, &name);
 
             if ( c<0 ) {
-                fprintf(stderr, "[fai_build_core] the last entry has no sequence\n");
+                hts_log_error("The last entry has no sequence");
                 goto fail;
             }
             if (c != '\n') while ( (c=bgzf_getc(bgzf))>=0 && c != '\n');
@@ -134,7 +138,7 @@ faidx_t *fai_build_core(BGZF *bgzf)
             offset = bgzf_utell(bgzf);
         } else {
             if (state == 3) {
-                fprintf(stderr, "[fai_build_core] inlined empty line is not allowed in sequence '%s'.\n", name.s);
+                hts_log_error("Inlined empty line is not allowed in sequence '%s'", name.s);
                 goto fail;
             }
             if (state == 2) state = 3;
@@ -144,7 +148,7 @@ faidx_t *fai_build_core(BGZF *bgzf)
                 if (isgraph(c)) ++l2;
             } while ( (c=bgzf_getc(bgzf))>=0 && c != '\n');
             if (state == 3 && l2) {
-                fprintf(stderr, "[fai_build_core] different line length in sequence '%s'.\n", name.s);
+                hts_log_error("Different line length in sequence '%s'", name.s);
                 goto fail;
             }
             ++l1; len += l2;
@@ -171,60 +175,76 @@ fail:
     return NULL;
 }
 
-void fai_save(const faidx_t *fai, FILE *fp)
-{
+static int fai_save(const faidx_t *fai, hFILE *fp) {
     khint_t k;
     int i;
+    char buf[96]; // Must be big enough for format below.
+
     for (i = 0; i < fai->n; ++i) {
         faidx1_t x;
         k = kh_get(s, fai->hash, fai->name[i]);
+        assert(k < kh_end(fai->hash));
         x = kh_value(fai->hash, k);
-#ifdef _WIN32
-        fprintf(fp, "%s\t%d\t%ld\t%d\t%d\n", fai->name[i], (int)x.len, (long)x.offset, (int)x.line_blen, (int)x.line_len);
-#else
-        fprintf(fp, "%s\t%d\t%lld\t%d\t%d\n", fai->name[i], (int)x.len, (long long)x.offset, (int)x.line_blen, (int)x.line_len);
-#endif
+        snprintf(buf, sizeof(buf),
+                 "\t%"PRId64"\t%"PRIu64"\t%"PRId32"\t%"PRId32"\n",
+                 x.len, x.offset, x.line_blen, x.line_len);
+        if (hputs(fai->name[i], fp) != 0) return -1;
+        if (hputs(buf, fp) != 0) return -1;
     }
+    return 0;
 }
 
-static faidx_t *fai_read(FILE *fp, const char *fname)
+static faidx_t *fai_read(hFILE *fp, const char *fname)
 {
     faidx_t *fai;
-    char *buf, *p;
-    int len, line_len, line_blen;
-#ifdef _WIN32
-    long offset;
-#else
-    long long offset;
-#endif
+    char *buf = NULL, *p;
+    int line_len, line_blen, n;
+    int64_t len;
+    uint64_t offset;
+    ssize_t l, lnum = 1;
+
     fai = (faidx_t*)calloc(1, sizeof(faidx_t));
+    if (!fai) return NULL;
+
     fai->hash = kh_init(s);
+    if (!fai->hash) goto fail;
+
     buf = (char*)calloc(0x10000, 1);
-    while (fgets(buf, 0x10000, fp)) {
-        for (p = buf; *p && isgraph(*p); ++p);
-        *p = 0; ++p;
-#ifdef _WIN32
-        sscanf(p, "%d%ld%d%d", &len, &offset, &line_blen, &line_len);
-#else
-        sscanf(p, "%d%lld%d%d", &len, &offset, &line_blen, &line_len);
-#endif
-        if (fai_insert_index(fai, buf, len, line_len, line_blen, offset) != 0) {
-            free(buf);
-            return NULL;
+    if (!buf) goto fail;
+
+    while ((l = hgetln(buf, 0x10000, fp)) > 0) {
+        for (p = buf; *p && !isspace_c(*p); ++p);
+        if (p - buf < l) {
+            *p = 0; ++p;
         }
+        n = sscanf(p, "%"SCNd64"%"SCNu64"%d%d", &len, &offset, &line_blen, &line_len);
+        if (n != 4) {
+            hts_log_error("Could not understand FAI %s line %zd", fname, lnum);
+            goto fail;
+        }
+        if (fai_insert_index(fai, buf, len, line_len, line_blen, offset) != 0) {
+            goto fail;
+        }
+        if (buf[l - 1] == '\n') ++lnum;
+    }
+
+    if (l < 0) {
+        hts_log_error("Error while reading %s: %s", fname, strerror(errno));
+        goto fail;
     }
     free(buf);
-    if (ferror(fp)) {
-        fprintf(stderr, "[fai_load] error while reading \"%s\": %s\n", fname, strerror(errno));
-        fai_destroy(fai);
-        return NULL;
-    }
     return fai;
+
+ fail:
+    free(buf);
+    fai_destroy(fai);
+    return NULL;
 }
 
 void fai_destroy(faidx_t *fai)
 {
     int i;
+    if (!fai) return;
     for (i = 0; i < fai->n; ++i) free(fai->name[i]);
     free(fai->name);
     kh_destroy(s, fai->hash);
@@ -232,206 +252,283 @@ void fai_destroy(faidx_t *fai)
     free(fai);
 }
 
-int fai_build(const char *fn)
+int fai_build3(const char *fn, const char *fnfai, const char *fngzi)
 {
-    char *str;
-    BGZF *bgzf;
-    FILE *fp;
-    faidx_t *fai;
-    str = (char*)calloc(strlen(fn) + 5, 1);
-    sprintf(str, "%s.fai", fn);
+    kstring_t fai_kstr = { 0, 0, NULL };
+    kstring_t gzi_kstr = { 0, 0, NULL };
+    BGZF *bgzf = NULL;
+    hFILE *fp = NULL;
+    faidx_t *fai = NULL;
+    int save_errno, res;
+
+    if (!fnfai) {
+        if (ksprintf(&fai_kstr, "%s.fai", fn) < 0) goto fail;
+        fnfai = fai_kstr.s;
+    }
+    if (!fngzi) {
+        if (ksprintf(&gzi_kstr, "%s.gzi", fn) < 0) goto fail;
+        fngzi = gzi_kstr.s;
+    }
+
     bgzf = bgzf_open(fn, "r");
     if ( !bgzf ) {
-        fprintf(stderr, "[fai_build] fail to open the FASTA file %s\n",fn);
-        free(str);
-        return -1;
-    }
-    if ( bgzf->is_compressed ) bgzf_index_build_init(bgzf);
-    fai = fai_build_core(bgzf);
-    if ( !fai )
-    {
-        if ( bgzf->is_compressed && bgzf->is_gzip ) fprintf(stderr,"Cannot index files compressed with gzip, please use bgzip\n");
-        bgzf_close(bgzf);
-        free(str);
-        return -1;
+        hts_log_error("Failed to open the FASTA file %s", fn);
+        goto fail;
     }
     if ( bgzf->is_compressed ) {
-        if (bgzf_index_dump(bgzf, fn, ".gzi") < 0) {
-            fprintf(stderr, "[fai_build] fail to make bgzf index %s.gzi\n", fn);
-            fai_destroy(fai); free(str);
-            return -1;
+        if (bgzf_index_build_init(bgzf) != 0) {
+            hts_log_error("Failed to allocate bgzf index");
+            goto fail;
         }
     }
-    if (bgzf_close(bgzf) < 0) {
-        fprintf(stderr, "[fai_build] Error on closing %s\n", fn);
-        fai_destroy(fai); free(str);
-        return -1;
+    fai = fai_build_core(bgzf);
+    if ( !fai ) {
+        if (bgzf->is_compressed && bgzf->is_gzip) {
+            hts_log_error("Cannot index files compressed with gzip, please use bgzip");
+        }
+        goto fail;
     }
-    fp = fopen(str, "wb");
+    if ( bgzf->is_compressed ) {
+        if (bgzf_index_dump(bgzf, fngzi, NULL) < 0) {
+            hts_log_error("Failed to make bgzf index %s", fngzi);
+            goto fail;
+        }
+    }
+    res = bgzf_close(bgzf);
+    bgzf = NULL;
+    if (res < 0) {
+        hts_log_error("Error on closing %s : %s", fn, strerror(errno));
+        goto fail;
+    }
+    fp = hopen(fnfai, "wb");
     if ( !fp ) {
-        fprintf(stderr, "[fai_build] fail to write FASTA index %s\n",str);
-        fai_destroy(fai); free(str);
-        return -1;
+        hts_log_error("Failed to open FASTA index %s : %s", fnfai, strerror(errno));
+        goto fail;
     }
-    fai_save(fai, fp);
-    fclose(fp);
-    free(str);
+    if (fai_save(fai, fp) != 0) {
+        hts_log_error("Failed to write FASTA index %s : %s", fnfai, strerror(errno));
+        goto fail;
+    }
+    if (hclose(fp) != 0) {
+        hts_log_error("Failed on closing FASTA index %s : %s", fnfai, strerror(errno));
+        goto fail;
+    }
+
+    free(fai_kstr.s);
+    free(gzi_kstr.s);
     fai_destroy(fai);
     return 0;
+
+ fail:
+    save_errno = errno;
+    free(fai_kstr.s);
+    free(gzi_kstr.s);
+    bgzf_close(bgzf);
+    fai_destroy(fai);
+    errno = save_errno;
+    return -1;
 }
 
-static FILE *download_and_open(const char *fn)
-{
-    const int buf_size = 1 * 1024 * 1024;
-    uint8_t *buf;
-    FILE *fp;
-    hFILE *fp_remote;
-    const char *url = fn;
-    const char *p;
-    int l = strlen(fn);
-    for (p = fn + l - 1; p >= fn; --p)
-        if (*p == '/') break;
-    fn = p + 1;
-
-    // First try to open a local copy
-    fp = fopen(fn, "r");
-    if (fp)
-        return fp;
-
-    // If failed, download from remote and open
-    fp_remote = hopen(url, "rb");
-    if (fp_remote == 0) {
-        fprintf(stderr, "[download_from_remote] fail to open remote file %s\n",url);
-        return NULL;
-    }
-    if ((fp = fopen(fn, "wb")) == 0) {
-        fprintf(stderr, "[download_from_remote] fail to create file in the working directory %s\n",fn);
-        hclose_abruptly(fp_remote);
-        return NULL;
-    }
-    buf = (uint8_t*)calloc(buf_size, 1);
-    while ((l = hread(fp_remote, buf, buf_size)) > 0)
-        fwrite(buf, 1, l, fp);
-    free(buf);
-    fclose(fp);
-    if (hclose(fp_remote) != 0)
-        fprintf(stderr, "[download_from_remote] fail to close remote file %s\n", url);
-
-    return fopen(fn, "r");
+int fai_build(const char *fn) {
+    return fai_build3(fn, NULL, NULL);
 }
 
-faidx_t *fai_load(const char *fn)
+faidx_t *fai_load3(const char *fn, const char *fnfai, const char *fngzi,
+                   int flags)
 {
-    char *str;
-    FILE *fp;
-    faidx_t *fai;
-    str = (char*)calloc(strlen(fn) + 5, 1);
-    sprintf(str, "%s.fai", fn);
+    kstring_t fai_kstr = { 0, 0, NULL };
+    kstring_t gzi_kstr = { 0, 0, NULL };
+    hFILE *fp = NULL;
+    faidx_t *fai = NULL;
+    int res;
 
-    if (hisremote(str))
-    {
-        fp = download_and_open(str);
-        if ( !fp )
-        {
-            fprintf(stderr, "[fai_load] failed to open remote FASTA index %s\n", str);
-            free(str);
-            return 0;
-        }
+    if (fn == NULL)
+        return NULL;
+
+    if (fnfai == NULL) {
+        if (ksprintf(&fai_kstr, "%s.fai", fn) < 0) goto fail;
+        fnfai = fai_kstr.s;
     }
-    else
-        fp = fopen(str, "rb");
+    if (fngzi == NULL) {
+        if (ksprintf(&gzi_kstr, "%s.gzi", fn) < 0) goto fail;
+        fngzi = gzi_kstr.s;
+    }
+
+    fp = hopen(fnfai, "rb");
 
     if (fp == 0) {
-        fprintf(stderr, "[fai_load] build FASTA index.\n");
-        if (fai_build(fn) < 0) {
-            free(str);
-            return 0;
+        if (!(flags & FAI_CREATE) || errno != ENOENT) {
+            hts_log_error("Failed to open FASTA index %s: %s", fnfai, strerror(errno));
+            goto fail;
         }
-        fp = fopen(str, "rb");
+
+        hts_log_info("Build FASTA index");
+
+        if (fai_build3(fn, fnfai, fngzi) < 0) {
+            goto fail;
+        }
+
+        fp = hopen(fnfai, "rb");
         if (fp == 0) {
-            fprintf(stderr, "[fai_load] failed to open FASTA index: %s\n", strerror(errno));
-            free(str);
-            return 0;
+            hts_log_error("Failed to open FASTA index %s: %s", fnfai, strerror(errno));
+            goto fail;
         }
     }
 
-    fai = fai_read(fp, str);
-    fclose(fp);
-    free(str);
+    fai = fai_read(fp, fnfai);
     if (fai == NULL) {
-        return NULL;
+        hts_log_error("Failed to read FASTA index %s", fnfai);
+        goto fail;
+    }
+
+    res = hclose(fp);
+    fp = NULL;
+    if (res < 0) {
+        hts_log_error("Failed on closing FASTA index %s : %s", fnfai, strerror(errno));
+        goto fail;
     }
 
     fai->bgzf = bgzf_open(fn, "rb");
     if (fai->bgzf == 0) {
-        fprintf(stderr, "[fai_load] fail to open FASTA file.\n");
-        return 0;
+        hts_log_error("Failed to open FASTA file %s", fn);
+        goto fail;
     }
-    if ( fai->bgzf->is_compressed==1 )
-    {
-        if ( bgzf_index_load(fai->bgzf, fn, ".gzi") < 0 )
-        {
-            fprintf(stderr, "[fai_load] failed to load .gzi index: %s[.gzi]\n", fn);
-            fai_destroy(fai);
-            return NULL;
+    if ( fai->bgzf->is_compressed==1 ) {
+        if ( bgzf_index_load(fai->bgzf, fngzi, NULL) < 0 ) {
+            hts_log_error("Failed to load .gzi index: %s", fngzi);
+            goto fail;
         }
     }
+    free(fai_kstr.s);
+    free(gzi_kstr.s);
     return fai;
+
+ fail:
+    if (fai) fai_destroy(fai);
+    if (fp) hclose_abruptly(fp);
+    free(fai_kstr.s);
+    free(gzi_kstr.s);
+    return NULL;
+}
+
+faidx_t *fai_load(const char *fn)
+{
+    return fai_load3(fn, NULL, NULL, FAI_CREATE);
+}
+
+static char *fai_retrieve(const faidx_t *fai, const faidx1_t *val,
+                          long beg, long end, int *len) {
+    char *s;
+    size_t l;
+    int c = 0;
+    int ret = bgzf_useek(fai->bgzf,
+                         val->offset
+                         + beg / val->line_blen * val->line_len
+                         + beg % val->line_blen, SEEK_SET);
+
+    if (ret < 0) {
+        *len = -1;
+        hts_log_error("Failed to retrieve block. (Seeking in a compressed, .gzi unindexed, file?)");
+        return NULL;
+    }
+
+    l = 0;
+    s = (char*)malloc((size_t) end - beg + 2);
+    if (!s) {
+        *len = -1;
+        return NULL;
+    }
+
+    while ( l < end - beg && (c=bgzf_getc(fai->bgzf))>=0 )
+        if (isgraph(c)) s[l++] = c;
+    if (c < 0) {
+        hts_log_error("Failed to retrieve block: %s",
+            c == -1 ? "unexpected end of file" : "error reading file");
+        free(s);
+        *len = -1;
+        return NULL;
+    }
+
+    s[l] = '\0';
+    *len = l < INT_MAX ? l : INT_MAX;
+    return s;
 }
 
 char *fai_fetch(const faidx_t *fai, const char *str, int *len)
 {
-    char *s;
-    int c, i, l, k, name_end;
+    char *s, *ep;
+    size_t i, l, k, name_end;
     khiter_t iter;
     faidx1_t val;
     khash_t(s) *h;
-    int beg, end;
+    long beg, end;
 
     beg = end = -1;
     h = fai->hash;
     name_end = l = strlen(str);
     s = (char*)malloc(l+1);
+    if (!s) {
+        *len = -1;
+        return NULL;
+    }
+
     // remove space
     for (i = k = 0; i < l; ++i)
-        if (!isspace(str[i])) s[k++] = str[i];
-    s[k] = 0; l = k;
+        if (!isspace_c(str[i])) s[k++] = str[i];
+    s[k] = 0;
+    name_end = l = k;
     // determine the sequence name
-    for (i = l - 1; i >= 0; --i) if (s[i] == ':') break; // look for colon from the end
-    if (i >= 0) name_end = i;
+    for (i = l; i > 0; --i) if (s[i - 1] == ':') break; // look for colon from the end
+    if (i > 0) name_end = i - 1;
     if (name_end < l) { // check if this is really the end
         int n_hyphen = 0;
         for (i = name_end + 1; i < l; ++i) {
             if (s[i] == '-') ++n_hyphen;
-            else if (!isdigit(s[i]) && s[i] != ',') break;
+            else if (!isdigit_c(s[i]) && s[i] != ',') break;
         }
         if (i < l || n_hyphen > 1) name_end = l; // malformated region string; then take str as the name
         s[name_end] = 0;
         iter = kh_get(s, h, s);
         if (iter == kh_end(h)) { // cannot find the sequence name
             iter = kh_get(s, h, str); // try str as the name
-            if (iter == kh_end(h)) {
-                *len = 0;
-            free(s); return 0;
-            } else s[name_end] = ':', name_end = l;
+            if (iter != kh_end(h)) {
+                s[name_end] = ':';
+                name_end = l;
+            }
         }
     } else iter = kh_get(s, h, str);
     if(iter == kh_end(h)) {
-        fprintf(stderr, "[fai_fetch] Warning - Reference %s not found in FASTA file, returning empty sequence\n", str);
+        hts_log_warning("Reference %s not found in FASTA file, returning empty sequence", str);
         free(s);
         *len = -2;
         return 0;
-    };
+    }
     val = kh_value(h, iter);
     // parse the interval
     if (name_end < l) {
+        int save_errno = errno;
+        errno = 0;
         for (i = k = name_end + 1; i < l; ++i)
             if (s[i] != ',') s[k++] = s[i];
         s[k] = 0;
-        beg = atoi(s + name_end + 1);
-        for (i = name_end + 1; i != k; ++i) if (s[i] == '-') break;
-        end = i < k? atoi(s + i + 1) : val.len;
+        if (s[name_end + 1] == '-') {
+            beg = 0;
+            i = name_end + 2;
+        } else {
+            beg = strtol(s + name_end + 1, &ep, 10);
+            for (i = ep - s; i < k;) if (s[i++] == '-') break;
+        }
+        end = i < k? strtol(s + i, &ep, 10) : val.len;
         if (beg > 0) --beg;
+        // Check for out of range numbers.  Only going to be a problem on
+        // 32-bit platforms with >2Gb sequence length.
+        if (errno == ERANGE && (uint64_t) val.len > LONG_MAX) {
+            hts_log_error("Positions in range %s are too large for this platform", s);
+            free(s);
+            *len = -2;
+            return NULL;
+        }
+        errno = save_errno;
     } else beg = 0, end = val.len;
     if (beg >= val.len) beg = val.len;
     if (end >= val.len) end = val.len;
@@ -439,20 +536,7 @@ char *fai_fetch(const faidx_t *fai, const char *str, int *len)
     free(s);
 
     // now retrieve the sequence
-    int ret = bgzf_useek(fai->bgzf, val.offset + beg / val.line_blen * val.line_len + beg % val.line_blen, SEEK_SET);
-    if ( ret<0 )
-    {
-        *len = -1;
-        fprintf(stderr, "[fai_fetch] Error: fai_fetch failed. (Seeking in a compressed, .gzi unindexed, file?)\n");
-        return NULL;
-    }
-    l = 0;
-    s = (char*)malloc(end - beg + 2);
-    while ( (c=bgzf_getc(fai->bgzf))>=0 && l < end - beg )
-        if (isgraph(c)) s[l++] = c;
-    s[l] = '\0';
-    *len = l;
-    return s;
+    return fai_retrieve(fai, &val, beg, end, len);
 }
 
 int faidx_fetch_nseq(const faidx_t *fai)
@@ -479,17 +563,15 @@ int faidx_seq_len(const faidx_t *fai, const char *seq)
 
 char *faidx_fetch_seq(const faidx_t *fai, const char *c_name, int p_beg_i, int p_end_i, int *len)
 {
-    int l, c;
     khiter_t iter;
     faidx1_t val;
-    char *seq=NULL;
 
     // Adjust position
     iter = kh_get(s, fai->hash, c_name);
     if (iter == kh_end(fai->hash))
     {
         *len = -2;
-        fprintf(stderr, "[fai_fetch_seq] The sequence \"%s\" not found\n", c_name);
+        hts_log_error("The sequence \"%s\" not found", c_name);
         return NULL;
     }
     val = kh_value(fai->hash, iter);
@@ -500,20 +582,7 @@ char *faidx_fetch_seq(const faidx_t *fai, const char *c_name, int p_beg_i, int p
     else if(val.len <= p_end_i) p_end_i = val.len - 1;
 
     // Now retrieve the sequence
-    int ret = bgzf_useek(fai->bgzf, val.offset + p_beg_i / val.line_blen * val.line_len + p_beg_i % val.line_blen, SEEK_SET);
-    if ( ret<0 )
-    {
-        *len = -1;
-        fprintf(stderr, "[fai_fetch_seq] Error: fai_fetch failed. (Seeking in a compressed, .gzi unindexed, file?)\n");
-        return NULL;
-    }
-    l = 0;
-    seq = (char*)malloc(p_end_i - p_beg_i + 2);
-    while ( (c=bgzf_getc(fai->bgzf))>=0 && l < p_end_i - p_beg_i + 1)
-        if (isgraph(c)) seq[l++] = c;
-    seq[l] = '\0';
-    *len = l;
-    return seq;
+    return fai_retrieve(fai, &val, p_beg_i, (long) p_end_i + 1, len);
 }
 
 int faidx_has_seq(const faidx_t *fai, const char *seq)
