@@ -1,6 +1,7 @@
-/*  sam.h -- SAM and BAM file I/O and manipulation.
-
-    Copyright (C) 2008, 2009, 2013-2014 Genome Research Ltd.
+/// @file htslib/sam.h
+/// High-level SAM/BAM/CRAM sequence file operations.
+/*
+    Copyright (C) 2008, 2009, 2013-2017 Genome Research Ltd.
     Copyright (C) 2010, 2012, 2013 Broad Institute.
 
     Author: Heng Li <lh3@sanger.ac.uk>
@@ -29,6 +30,10 @@ DEALINGS IN THE SOFTWARE.  */
 #include <stdint.h>
 #include "hts.h"
 
+#ifdef __cplusplus
+extern "C" {
+#endif
+
 /**********************
  *** SAM/BAM header ***
  **********************/
@@ -75,7 +80,11 @@ typedef struct {
 
 #define bam_cigar_op(c) ((c)&BAM_CIGAR_MASK)
 #define bam_cigar_oplen(c) ((c)>>BAM_CIGAR_SHIFT)
-#define bam_cigar_opchr(c) (BAM_CIGAR_STR[bam_cigar_op(c)])
+// Note that BAM_CIGAR_STR is padded to length 16 bytes below so that
+// the array look-up will not fall off the end.  '?' is chosen as the
+// padding character so it's easy to spot if one is emitted, and will
+// result in a parsing failure (in sam_parse1(), at least) if read.
+#define bam_cigar_opchr(c) (BAM_CIGAR_STR "??????" [bam_cigar_op(c)])
 #define bam_cigar_gen(l, o) ((l)<<BAM_CIGAR_SHIFT|(o))
 
 /* bam_cigar_type returns a bit flag with:
@@ -136,6 +145,7 @@ typedef struct {
  @field  qual    mapping quality
  @field  l_qname length of the query name
  @field  flag    bitwise flag
+ @field  l_extranul length of extra NULs between qname & cigar (for alignment)
  @field  n_cigar number of CIGAR operations
  @field  l_qseq  length of the query sequence (read)
  @field  mtid    chromosome ID of next read in template, defined by bam_hdr_t
@@ -144,8 +154,13 @@ typedef struct {
 typedef struct {
     int32_t tid;
     int32_t pos;
-    uint32_t bin:16, qual:8, l_qname:8;
-    uint32_t flag:16, n_cigar:16;
+    uint16_t bin;
+    uint8_t qual;
+    uint8_t l_qname;
+    uint16_t flag;
+    uint8_t unused1;
+    uint8_t l_extranul;
+    uint32_t n_cigar;
     int32_t l_qseq;
     int32_t mtid;
     int32_t mpos;
@@ -161,7 +176,9 @@ typedef struct {
 
  @discussion Notes:
 
- 1. qname is zero tailing and core.l_qname includes the tailing '\0'.
+ 1. qname is terminated by one to four NULs, so that the following
+ cigar data is 32-bit aligned; core.l_qname includes these trailing NULs,
+ while core.l_extranul counts the excess NULs (so 0 <= l_extranul <= 3).
  2. l_qseq is calculated from the total length of an alignment block
  on reading or from CIGAR.
  3. cigar data is encoded 4 bytes per CIGAR operation.
@@ -169,7 +186,8 @@ typedef struct {
  */
 typedef struct {
     bam1_core_t core;
-    int l_data, m_data;
+    int l_data;
+    uint32_t m_data;
     uint8_t *data;
 #ifndef BAM_NO_ID
     uint64_t id;
@@ -245,25 +263,21 @@ typedef struct {
  *** Exported functions ***
  **************************/
 
-#ifdef __cplusplus
-extern "C" {
-#endif
-
     /***************
      *** BAM I/O ***
      ***************/
 
     bam_hdr_t *bam_hdr_init(void);
     bam_hdr_t *bam_hdr_read(BGZF *fp);
-    int bam_hdr_write(BGZF *fp, const bam_hdr_t *h);
+    int bam_hdr_write(BGZF *fp, const bam_hdr_t *h) HTS_RESULT_USED;
     void bam_hdr_destroy(bam_hdr_t *h);
     int bam_name2id(bam_hdr_t *h, const char *ref);
     bam_hdr_t* bam_hdr_dup(const bam_hdr_t *h0);
 
     bam1_t *bam_init1(void);
     void bam_destroy1(bam1_t *b);
-    int bam_read1(BGZF *fp, bam1_t *b);
-    int bam_write1(BGZF *fp, const bam1_t *b);
+    int bam_read1(BGZF *fp, bam1_t *b) HTS_RESULT_USED;
+    int bam_write1(BGZF *fp, const bam1_t *b) HTS_RESULT_USED;
     bam1_t *bam_copy1(bam1_t *bdst, const bam1_t *bsrc);
     bam1_t *bam_dup1(const bam1_t *bsrc);
 
@@ -297,13 +311,44 @@ extern "C" {
     #define bam_itr_querys(idx, hdr, region) sam_itr_querys(idx, hdr, region)
     #define bam_itr_next(htsfp, itr, r) hts_itr_next((htsfp)->fp.bgzf, (itr), (r), 0)
 
-    // Load .csi or .bai BAM index file.
-    #define bam_index_load(fn) hts_idx_load((fn), HTS_FMT_BAI)
+// Load/build .csi or .bai BAM index file.  Does not work with CRAM.
+// It is recommended to use the sam_index_* functions below instead.
+#define bam_index_load(fn) hts_idx_load((fn), HTS_FMT_BAI)
+#define bam_index_build(fn, min_shift) (sam_index_build((fn), (min_shift)))
 
-    int bam_index_build(const char *fn, int min_shift);
+/// Load a BAM (.csi or .bai) or CRAM (.crai) index file
+/** @param fp  File handle of the data file whose index is being opened
+    @param fn  BAM/CRAM/etc filename to search alongside for the index file
+    @return  The index, or NULL if an error occurred.
+*/
+hts_idx_t *sam_index_load(htsFile *fp, const char *fn);
 
-    // Load BAM (.csi or .bai) or CRAM (.crai) index file.
-    hts_idx_t *sam_index_load(htsFile *fp, const char *fn);
+/// Load a specific BAM (.csi or .bai) or CRAM (.crai) index file
+/** @param fp     File handle of the data file whose index is being opened
+    @param fn     BAM/CRAM/etc data file filename
+    @param fnidx  Index filename, or NULL to search alongside @a fn
+    @return  The index, or NULL if an error occurred.
+*/
+hts_idx_t *sam_index_load2(htsFile *fp, const char *fn, const char *fnidx);
+
+/// Generate and save an index file
+/** @param fn        Input BAM/etc filename, to which .csi/etc will be added
+    @param min_shift Positive to generate CSI, or 0 to generate BAI
+    @return  0 if successful, or negative if an error occurred (usually -1; or
+             -2: opening fn failed; -3: format not indexable; -4:
+             failed to create and/or save the index)
+*/
+int sam_index_build(const char *fn, int min_shift) HTS_RESULT_USED;
+
+/// Generate and save an index to a specific file
+/** @param fn        Input BAM/CRAM/etc filename
+    @param fnidx     Output filename, or NULL to add .bai/.csi/etc to @a fn
+    @param min_shift Positive to generate CSI, or 0 to generate BAI
+    @return  0 if successful, or negative if an error occurred (see
+             sam_index_build for error codes)
+*/
+int sam_index_build2(const char *fn, const char *fnidx, int min_shift) HTS_RESULT_USED;
+int sam_index_build3(const char *fn, const char *fnidx, int min_shift, int nthreads) HTS_RESULT_USED;
 
     #define sam_itr_destroy(iter) hts_itr_destroy(iter)
     hts_itr_t *sam_itr_queryi(const hts_idx_t *idx, int tid, int beg, int end);
@@ -315,36 +360,131 @@ extern "C" {
      ***************/
 
     #define sam_open(fn, mode) (hts_open((fn), (mode)))
+    #define sam_open_format(fn, mode, fmt) (hts_open_format((fn), (mode), (fmt)))
     #define sam_close(fp) hts_close(fp)
 
     int sam_open_mode(char *mode, const char *fn, const char *format);
 
+    // A version of sam_open_mode that can handle ,key=value options.
+    // The format string is allocated and returned, to be freed by the caller.
+    // Prefix should be "r" or "w",
+    char *sam_open_mode_opts(const char *fn,
+                             const char *mode,
+                             const char *format);
+
     typedef htsFile samFile;
     bam_hdr_t *sam_hdr_parse(int l_text, const char *text);
     bam_hdr_t *sam_hdr_read(samFile *fp);
-    int sam_hdr_write(samFile *fp, const bam_hdr_t *h);
+    int sam_hdr_write(samFile *fp, const bam_hdr_t *h) HTS_RESULT_USED;
 
-    int sam_parse1(kstring_t *s, bam_hdr_t *h, bam1_t *b);
-    int sam_format1(const bam_hdr_t *h, const bam1_t *b, kstring_t *str);
-    int sam_read1(samFile *fp, bam_hdr_t *h, bam1_t *b);
-    int sam_write1(samFile *fp, const bam_hdr_t *h, const bam1_t *b);
+    int sam_parse1(kstring_t *s, bam_hdr_t *h, bam1_t *b) HTS_RESULT_USED;
+    int sam_format1(const bam_hdr_t *h, const bam1_t *b, kstring_t *str) HTS_RESULT_USED;
+    int sam_read1(samFile *fp, bam_hdr_t *h, bam1_t *b) HTS_RESULT_USED;
+    int sam_write1(samFile *fp, const bam_hdr_t *h, const bam1_t *b) HTS_RESULT_USED;
 
     /*************************************
      *** Manipulating auxiliary fields ***
      *************************************/
 
-    uint8_t *bam_aux_get(const bam1_t *b, const char tag[2]);
-    int32_t bam_aux2i(const uint8_t *s);
-    double bam_aux2f(const uint8_t *s);
-    char bam_aux2A(const uint8_t *s);
-    char *bam_aux2Z(const uint8_t *s);
+/// Return a pointer to an aux record
+/** @param b   Pointer to the bam record
+    @param tag Desired aux tag
+    @return Pointer to the tag data, or NULL if tag is not present or on error
+    If the tag is not present, this function returns NULL and sets errno to
+    ENOENT.  If the bam record's aux data is corrupt (either a tag has an
+    invalid type, or the last record is incomplete) then errno is set to
+    EINVAL and NULL is returned.
+ */
+uint8_t *bam_aux_get(const bam1_t *b, const char tag[2]);
 
-    void bam_aux_append(bam1_t *b, const char tag[2], char type, int len, uint8_t *data);
-    int bam_aux_del(bam1_t *b, uint8_t *s);
+/// Get an integer aux value
+/** @param s Pointer to the tag data, as returned by bam_aux_get()
+    @return The value, or 0 if the tag was not an integer type
+    If the tag is not an integer type, errno is set to EINVAL.  This function
+    will not return the value of floating-point tags.
+*/
+int64_t bam_aux2i(const uint8_t *s);
 
-#ifdef __cplusplus
-}
-#endif
+/// Get an integer aux value
+/** @param s Pointer to the tag data, as returned by bam_aux_get()
+    @return The value, or 0 if the tag was not an integer type
+    If the tag is not an numeric type, errno is set to EINVAL.  The value of
+    integer flags will be returned cast to a double.
+*/
+double bam_aux2f(const uint8_t *s);
+
+/// Get a character aux value
+/** @param s Pointer to the tag data, as returned by bam_aux_get().
+    @return The value, or 0 if the tag was not a character ('A') type
+    If the tag is not a character type, errno is set to EINVAL.
+*/
+char bam_aux2A(const uint8_t *s);
+
+/// Get a string aux value
+/** @param s Pointer to the tag data, as returned by bam_aux_get().
+    @return Pointer to the string, or NULL if the tag was not a string type
+    If the tag is not a string type ('Z' or 'H'), errno is set to EINVAL.
+*/
+char *bam_aux2Z(const uint8_t *s);
+
+/// Get the length of an array-type ('B') tag
+/** @param s Pointer to the tag data, as returned by bam_aux_get().
+    @return The length of the array, or 0 if the tag is not an array type.
+    If the tag is not an array type, errno is set to EINVAL.
+ */
+uint32_t bam_auxB_len(const uint8_t *s);
+
+/// Get an integer value from an array-type tag
+/** @param s   Pointer to the tag data, as returned by bam_aux_get().
+    @param idx 0-based Index into the array
+    @return The idx'th value, or 0 on error.
+    If the array is not an integer type, errno is set to EINVAL.  If idx
+    is greater than or equal to  the value returned by bam_auxB_len(s),
+    errno is set to ERANGE.  In both cases, 0 will be returned.    
+ */
+int64_t bam_auxB2i(const uint8_t *s, uint32_t idx);
+
+/// Get a floating-point value from an array-type tag
+/** @param s   Pointer to the tag data, as returned by bam_aux_get().
+    @param idx 0-based Index into the array
+    @return The idx'th value, or 0.0 on error.
+    If the array is not a numeric type, errno is set to EINVAL.  This can
+    only actually happen if the input record has an invalid type field.  If
+    idx is greater than or equal to  the value returned by bam_auxB_len(s),
+    errno is set to ERANGE.  In both cases, 0.0 will be returned.
+ */
+double bam_auxB2f(const uint8_t *s, uint32_t idx);
+
+/// Append tag data to a bam record
+/* @param b    The bam record to append to.
+   @param tag  Tag identifier
+   @param type Tag data type
+   @param len  Length of the data in bytes
+   @param data The data to append
+   @return 0 on success; -1 on failure.
+If there is not enough space to store the additional tag, errno is set to
+ENOMEM.  If the type is invalid, errno may be set to EINVAL.  errno is
+also set to EINVAL if the bam record's aux data is corrupt.
+*/
+int bam_aux_append(bam1_t *b, const char tag[2], char type, int len, const uint8_t *data);
+
+/// Delete tag data from a bam record
+/* @param b The bam record to update
+   @param s Pointer to the tag to delete, as returned by bam_aux_get().
+   @return 0 on success; -1 on failure
+   If the bam record's aux data is corrupt, errno is set to EINVAL and this
+   function returns -1;
+*/
+int bam_aux_del(bam1_t *b, uint8_t *s);
+
+/// Update a string-type tag
+/* @param b    The bam record to update
+   @param tag  Tag identifier
+   @param len  The length of the new string
+   @param data The new string
+   @return 0 on success, -1 on failure
+*/
+int bam_aux_update_str(bam1_t *b, const char tag[2], int len, const char *data);
 
 /**************************
  *** Pileup and Mpileup ***
@@ -352,6 +492,21 @@ extern "C" {
 
 #if !defined(BAM_NO_PILEUP)
 
+/*! @typedef
+ @abstract Generic pileup 'client data'.
+
+ @discussion The pileup iterator allows setting a constructor and
+ destructor function, which will be called every time a sequence is
+ fetched and discarded.  This permits caching of per-sequence data in
+ a tidy manner during the pileup process.  This union is the cached
+ data to be manipulated by the "client" (the caller of pileup).
+*/
+typedef union {
+    void *p;
+    int64_t i;
+    double f;
+} bam_pileup_cd;
+
 /*! @typedef
  @abstract Structure for one alignment covering the pileup position.
  @field  b          pointer to the alignment
@@ -375,6 +530,7 @@ typedef struct {
     int32_t qpos;
     int indel, level;
     uint32_t is_del:1, is_head:1, is_tail:1, is_refskip:1, aux:28;
+    bam_pileup_cd cd; // generic per-struct data, owned by caller.
 } bam_pileup1_t;
 
 typedef int (*bam_plp_auto_f)(void *data, bam1_t *b);
@@ -385,10 +541,6 @@ typedef struct __bam_plp_t *bam_plp_t;
 struct __bam_mplp_t;
 typedef struct __bam_mplp_t *bam_mplp_t;
 
-#ifdef __cplusplus
-extern "C" {
-#endif
-
     /**
      *  bam_plp_init() - sets an iterator over multiple
      *  @func:      see mplp_func in bam_plcmd.c in samtools for an example. Expected return
@@ -403,6 +555,19 @@ extern "C" {
     void bam_plp_set_maxcnt(bam_plp_t iter, int maxcnt);
     void bam_plp_reset(bam_plp_t iter);
 
+    /**
+     *  bam_plp_constructor() - sets a callback to initialise any per-pileup1_t fields.
+     *  @plp:       The bam_plp_t initialised using bam_plp_init.
+     *  @func:      The callback function itself.  When called, it is given the
+     *              data argument (specified in bam_plp_init), the bam structure and
+     *              a pointer to a locally allocated bam_pileup_cd union.  This union
+     *              will also be present in each bam_pileup1_t created.
+     */
+    void bam_plp_constructor(bam_plp_t plp,
+                             int (*func)(void *data, const bam1_t *b, bam_pileup_cd *cd));
+    void bam_plp_destructor(bam_plp_t plp,
+                            int (*func)(void *data, const bam1_t *b, bam_pileup_cd *cd));
+
     bam_mplp_t bam_mplp_init(int n, bam_plp_auto_f func, void **data);
     /**
      *  bam_mplp_init_overlaps() - if called, mpileup will detect overlapping
@@ -416,11 +581,24 @@ extern "C" {
     void bam_mplp_destroy(bam_mplp_t iter);
     void bam_mplp_set_maxcnt(bam_mplp_t iter, int maxcnt);
     int bam_mplp_auto(bam_mplp_t iter, int *_tid, int *_pos, int *n_plp, const bam_pileup1_t **plp);
+    void bam_mplp_reset(bam_mplp_t iter);
+    void bam_mplp_constructor(bam_mplp_t iter,
+                              int (*func)(void *data, const bam1_t *b, bam_pileup_cd *cd));
+    void bam_mplp_destructor(bam_mplp_t iter,
+                             int (*func)(void *data, const bam1_t *b, bam_pileup_cd *cd));
+
+#endif // ~!defined(BAM_NO_PILEUP)
+
+
+/***********************************
+ * BAQ calculation and realignment *
+ ***********************************/
+
+int sam_cap_mapq(bam1_t *b, const char *ref, int ref_len, int thres);
+int sam_prob_realn(bam1_t *b, const char *ref, int ref_len, int flag);
 
 #ifdef __cplusplus
 }
 #endif
 
-#endif // ~!defined(BAM_NO_PILEUP)
-
 #endif
